#include <stdio.h>
/*#include "/usr/local/express/include/express.h"*/
#include "smfa.h"
#include "decl.h"
#include <math.h>

#define EPSILON   0.000000        /* Delta_Energy least acceptable */

#define DEBUG 0
#define DEBUG2 0
#define TIMING 1
#define DEBUG3 1
#define is_R_not_constant 1
#define RSB 1

main(argc, argv)
int argc;
char **argv;
{
/*
c#######################################################################
c
c     Variable Declarations.
c
c#######################################################################
*/

	int THRESHOLD = 4;
	double *pcg;	 				/* Processor Comm Graph row vector */
	double gamma_p;				/* Load factor in mf calculation */
	double **lambda;
	double **spin;
	int num_vertices;
	double delta_H = 0.0, R;
	int element;
	int type4 = 34;
	double spin_new[NPROC];
	double  temp;
	int in_pass=0, max_iterations;
	vertex_t *vertex;
	double *mean_field;
	double K;
	FILE *fin;
	int equilibrium=0, out_pass=0;
	int cnt_in_mvs=0, total_moves=0;
	float t1,t2;
	float etime();
    void srand48();
    long seedval;
	int *map, j;
	int ref_iter;


/*
c@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@
c
c     Beginning of executable code...
c
c@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@
*/


	/* Cmd Line validation */
	if ( argc != 6 )      /* Make it 3 for ver2 */
		exit(99);         /* Illegal Number Of Parameters */
	/* Right now:
	   ARGV1 =>  Graph name to be read
	   ARGV2 =>  value of R to be read
	   ARGV3 =>  value of temp to be read
	   ARGV4 =>  RSB file name with initial mapping
	   ARGV5 =>  incremental points' file name      
     */

	/* read the graph to be partitioned */
   fin = open_file( argv[1], READ);


   num_vertices = read_max_vertices(fin);
   vertex = MEM(vertex_t, num_vertices);

#if DEBUG2
    printf("\n num_vertices:%d\n",num_vertices);
#endif


   if ( vertex == NULL )
    exit(10);   /* Out Of memory Condition */

   if ( read_graph(fin, vertex, num_vertices) < 0)
      exit(2);          /* Some error in Reading Graph */
	fclose(fin);

#if DEBUG2
    printf("\n After read graph\n");
#endif

	/* Set the initial temperature to 5.0 degrees */
	/*temp = TEMP0;*/
    sscanf(argv[3],"%lf",&temp);

	/* For timing purpose read no of iterations to be timed:*/
    /* sscanf(argv[4],"%d",&ref_iter); */

#if RSB
    fin = open_file(argv[4],READ);
    fscanf(fin,"%d",&num_vertices);
    for (j=0;j<num_vertices;j++) 
        fscanf(fin,"%d",&map[j]);
    fclose(fin);
#endif RSB


	/* Initialize the Spin matrix's local vector */
	spin = init_spin_matrix(num_vertices, map);

#if DEBUG3
	map = generate_solution( spin, num_vertices );
    printf("Initial ");
    find_solution_quality( map, num_vertices, vertex );
#endif

#if DEBUG2
    printf("\n After init spin\n");
#endif

	/*  evaluate lambda[i][p] */
	lambda = eval_lambda( num_vertices, vertex, spin);

#if DEBUG2
    printf("\n After lambda\n");
#endif

#if is_R_not_constant
	sscanf(argv[2],"%lf",&R);  /* 4.202805e-05;*/
#endif

#if DEBUG3
    printf("\n  After R is %f\n", R);
#endif

	max_iterations = num_vertices;

    /* Initialize the random no. generator */
    seedval = 280;
    srand48( seedval );

    /* Timing using express timer, after file i/o */
#if TIMING
    t1 = etime();  /* This fixes the reference point for timing */
#endif

	while (  temp > (TEMP0/1000.0)  ) {
		out_pass++;

		/* inner While Loop of MFA */
		while ( (! equilibrium )&&(cnt_in_mvs < max_iterations) ) {
			cnt_in_mvs++;
			total_moves++;

#if 0
			/* To print the time for ref_iter no. of moves */
			if (total_moves == ref_iter) {
				t2 = etime();
				printf("\nRef_iter: %d,Time/iteration : %f\n",
						  ref_iter,(t2-t1)/ref_iter);
			}
#endif      

			element = perturb(num_vertices);

			mean_field = eval_mean_field( element, lambda[element]
                                          , R, spin, vertex, num_vertices );

#if DEBUG2
			printf(" After Mean Field %f\n", mean_field[0]);
#endif

			for ( j=0; j < NPROC; j++) {
				spin_new[j] = tanh( mean_field[j] / temp);
				delta_H = delta_H + mean_field[j]*
									( spin_new[j] - spin[element][j] );
			}

#if DEBUG
				printf("Delta H %f Temp %f IN PASS %d Net Mvs %d\n",
						delta_H, temp, in_pass, cnt_in_mvs);
#endif

			/* Update Spin averages */
            update_spin( element, spin, spin_new );
 
			if ( (fabs(delta_H) == EPSILON)||( delta_H > 0.0 ))
				in_pass++;
			else
				in_pass = 0;

			if ( in_pass == THRESHOLD )
				equilibrium = 1;
#if DEBUG2
	map = generate_solution( spin, num_vertices );
	printf("%d. ", total_moves);
    find_solution_quality( map, num_vertices, vertex );
#endif

		}   /* End Of Inner While Loop */

		in_pass     = 0;  /* reset in_pass count here */
		equilibrium = 0;
		cnt_in_mvs  = 0;

		if ( temp > ( TEMP0 / 1.5) )  /* For slow cooling schedule */
			K = 0.95;
		else {
			K = 0.85;
			max_iterations = num_vertices / 4;
			THRESHOLD = THRESHOLD/4;
		}

		temp = K * temp;
	}  /* End Of Outter While loop */

#if TIMING
    t2 = etime();
    printf(" Elapsed Time %f\n", t2-t1);
#endif


#if DEBUG3
	printf("\n Total No Of Outter Iterations %d Total Moves %d \n",
			out_pass, total_moves);
#endif

	map = generate_solution( spin, num_vertices );
    find_solution_quality( map, num_vertices, vertex );

}   /* End Of Main Program */



void printv( vertex, i)
vertex_t *vertex;
int i;
{

    int j;

    printf(" \n");
    for( j=0; j < vertex[i].deg; j ++ ) {

        printf(" Vertex Lsn %d (%d). Nabor: %d EgWt %f\n",
                 i, j,vertex[i].nabor[j], vertex[i].egwt[j]);
    }

}


/*------------------------------------------------------------*/

int *dec2bin(dec, numOfbits)
int dec;
int numOfbits;
{
	int i;
	int *binNum;
	if ((binNum = MEM(int, numOfbits))==NULL)
		printf("\n Memory allocation error \n");
	for(i=0; i < numOfbits; i++) {
		binNum[i] = dec % 2;
		dec = dec/2;
	}
	return binNum;
}
