#include <stdio.h>
/*#include "/usr/local/express/include/express.h"*/
#include <math.h>
#include "smfa.h"
#include "decl.h"

#define TEST_MAIN 0
#define DEBUG 0


#if TEST_MAIN

main()
{

	double *lambda;
    vertex_t *vertex;
    double *spin;
    int num_vertices;

    lambda = eval_lambda( num_vertices, vertex, spin);

}


#endif


/* This function computes the column vector lambda[i] */

double **eval_lambda( num_vertices, vertex, spin)
int num_vertices;
vertex_t *vertex;
double **spin;
{
	int i, j, k;
	double **lambda;

    lambda = (double **)calloc(num_vertices, sizeof(double *));

	if ( lambda ==  NULL )
		exit(10);       /* Out Of memory Condition */

    for ( i=0; i < num_vertices; i++ )
		if ( (lambda[i] = MEM(double,NPROC))==NULL)
			exit(10);


	for ( k=0; k < NPROC; k++) {
		for ( i=0; i < num_vertices; i++ ) {
       		lambda[i][k] = 0.0;
    		for ( j=0; j < vertex[i].deg; j++ )
        		lambda[i][k] = lambda[i][k] + ( vertex[i].egwt[j]
                        * spin[vertex[i].nabor[j]][k] );
		}
	}

	return lambda;

}

double eval_coeff_R( num_vertices,lambda, spin, vertex )
int num_vertices;
double **lambda;
double **spin;
vertex_t *vertex;
{
    int i, j;
    double R[2];
    double *sum_lambda;
    double gamma_p=0.0, sum2=0.0, factor;
	double R_prob;


    if ( (sum_lambda = MEM(double,num_vertices)) == NULL )
        exit(10);       /* Out Of memory Condition */


    for ( i=0; i < num_vertices; i++ ) {
        sum_lambda[i] = 0.0;
        for ( j=0; j < NPROC; j++ )
            sum_lambda[i] = sum_lambda[i] + lambda[i][j];
    }


    R[0] = 0.0;
    for ( j=0; j < NPROC; j++ )
        for ( i=0; i < num_vertices; i++ )
            R[0] = R[0] + spin[i][j] * (sum_lambda[i] - lambda[i][j] );

    R[1] =0.0;
    for ( j=0; j < NPROC; j++ ) {
        for ( i=0; i < num_vertices; i++ ) {
            factor = vertex[i].wt * spin[i][j];
            gamma_p = gamma_p + factor;
            sum2 = sum2 + factor * factor;
        }
        R[1] = R[1] + gamma_p*gamma_p - sum2;
    }

    if ( R[1] == 0.0 )
        exit(20);           /* Divide By Zero */
    else {
		R_prob = R[0]/(R[1]*NPROC);
		if ( R_prob < 0.0 )
			R_prob = - R_prob;
        return (R_prob);
	}

}



int
addf(p1, p2, size)
  double          *p1, *p2;
  int             size;
{
  *p1 += *p2;
  return (0);
}


double eval_psi( mynode, i, spin, vertex , num_vertices)
int mynode;
int i;
double **spin;
vertex_t *vertex;
int num_vertices;
{
	int j, q, k;
	double tmp_spin = 1.0;
	double psi_ik = 0;

	for( j=0; j < num_vertices; j++ ) {

		if ( j != i) {

			for( q=0; q < NPROC; q++) {
				if ( q != mynode )
					tmp_spin = tmp_spin *(1+spin[i][q]* spin[j][q] );
			}
			psi_ik = psi_ik + vertex[j].wt*spin[j][mynode]*tmp_spin;
		}
	}

#if DEBUG
    printf("\n Psi_ik %f\n",psi_ik);
#endif

	return ( psi_ik * vertex[i].wt );
		
}


double *eval_mean_field( element, lambda_i, R,  spin, vertex, num_vertices)
int element;
double *lambda_i;
double R;
double **spin;
vertex_t *vertex;
int num_vertices;
{

	double  psi_p;
	double *mean_field;
	int j;

    mean_field    = MEM(double,NPROC);

	if ( mean_field == NULL )
		exit(10);

    for ( j=0; j < NPROC; j++) {
		psi_p = eval_psi( j, element, spin, vertex, num_vertices );
		mean_field[j] = lambda_i[j] - ( R*psi_p); 
	}

    return ( mean_field );

}


double *eval_abc( mean_field ,  temp, spin_ip )
double mean_field;
double temp;
double spin_ip;
{

	double *abc_ip;

	if ( (abc_ip = MEM(double,3)) == NULL)
		exit(10);		/* Out Of memory condition */


	if ( temp == 0.0 )
		exit(22);       /* Exp will bomb here */

	/* Compute a_ip */
	abc_ip[0] = (double) exp( ((double) mean_field) / ( (double) temp) );

	/* Compute b_ip */
	abc_ip[1] = abc_ip[0]*mean_field;

	/* Compute c_ip */
	abc_ip[2] = mean_field * spin_ip;

	return abc_ip;

}

